# Generated by Django 2.1.3 on 2018-11-18 12:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0012_auto_20181117_1745'),
    ]

    operations = [
        migrations.AddField(
            model_name='cwrexport',
            name='_work_count',
            field=models.IntegerField(editable=False, null=True),
        ),
    ]



# Generated by Django 2.1 on 2018-09-11 10:17

from django.db import migrations, models
import music_publisher.models


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0006_auto_20180801_2130'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='albumcd',
            options={'ordering': ('album_title', 'cd_identifier'), 'verbose_name': 'Album and/or Library CD', 'verbose_name_plural': ' Albums and Library CDs'},
        ),
        migrations.AlterModelOptions(
            name='artist',
            options={'verbose_name': 'Performing Artist', 'verbose_name_plural': ' Performing Artists'},
        ),
        migrations.AlterModelOptions(
            name='cwrexport',
            options={'ordering': ('-id',), 'verbose_name': 'CWR Export', 'verbose_name_plural': 'CWR Exports'},
        ),
        migrations.AlterModelOptions(
            name='work',
            options={'verbose_name_plural': '  Works'},
        ),
        migrations.AlterModelOptions(
            name='workacknowledgement',
            options={'verbose_name': 'Registration Acknowledgement'},
        ),
        migrations.AlterModelOptions(
            name='writer',
            options={'ordering': ('last_name', 'first_name', 'ipi_name'), 'verbose_name_plural': ' Writers'},
        ),
        migrations.AlterModelOptions(
            name='writerinwork',
            options={'ordering': ('-controlled', 'writer__last_name', 'writer__first_name'), 'verbose_name_plural': 'Writers in Work'},
        ),
        migrations.AlterField(
            model_name='work',
            name='last_change',
            field=models.DateTimeField(editable=False, null=True, verbose_name='Last Edited'),
        ),
        migrations.AlterField(
            model_name='writer',
            name='ipi_name',
            field=models.CharField(blank=True, help_text='Required for a controlled writer', max_length=11, null=True, unique=True, validators=[music_publisher.models.CWRFieldValidator('writer_ipi_name')], verbose_name='IPI Name #'),
        ),
        migrations.AlterField(
            model_name='writer',
            name='pr_society',
            field=models.CharField(blank=True, choices=[('101', 'SOCAN, Canada'), ('088', 'CMRRA, Canada'), ('058', 'SACEM, France'), ('068', 'SDRM, France'), ('035', 'GEMA, Germany'), ('074', 'SIAE, Italy'), ('052', 'PRS, United Kingdom'), ('044', 'MCPS, United Kingdom'), ('010', 'ASCAP, United States'), ('021', 'BMI, United States'), ('071', 'SESAC Inc., United States'), ('034', 'HFA, United States')], help_text='Required for a controlled writer', max_length=3, null=True, validators=[music_publisher.models.CWRFieldValidator('writer_pr_society')], verbose_name='Performing Rights Society'),
        ),
    ]



from django.apps import AppConfig


class MusicPublisherConfig(AppConfig):

    """Configuration for Music Publisher app.

    Attributes:
        label (str): app label
        name (str): app name
        verbose_name (str): app verbose name
    """

    name = 'music_publisher'
    label = 'music_publisher'
    verbose_name = 'Music Publisher'



# Generated by Django 2.1 on 2018-09-12 12:17

import django.core.validators
from django.db import migrations, models
import music_publisher.models


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0008_auto_20180911_1055'),
    ]

    operations = [
        migrations.AddField(
            model_name='albumcd',
            name='album_label',
            field=models.CharField(blank=True, default='FOO BAR MUSIC', max_length=60, validators=[music_publisher.models.CWRFieldValidator('first_album_label')]),
        ),
        migrations.AlterField(
            model_name='albumcd',
            name='album_title',
            field=models.CharField(blank=True, default='', max_length=60, unique=True, validators=[music_publisher.models.CWRFieldValidator('first_album_title')]),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='workacknowledgement',
            name='society_code',
            field=models.CharField(choices=[('055', 'SABAM, Belgium'), ('101', 'SOCAN, Canada'), ('088', 'CMRRA, Canada'), ('040', 'KODA, Denmark'), ('089', 'TEOSTO, Finland'), ('058', 'SACEM, France'), ('035', 'GEMA, Germany'), ('074', 'SIAE, Italy'), ('023', 'BUMA, Netherlands'), ('078', 'STEMRA, Netherlands'), ('090', 'TONO, Norway'), ('079', 'STIM, Sweden'), ('052', 'PRS, United Kingdom'), ('044', 'MCPS, United Kingdom'), ('010', 'ASCAP, United States'), ('021', 'BMI, United States'), ('071', 'SESAC Inc., United States'), ('034', 'HFA, United States')], max_length=3, verbose_name='Society'),
        ),
        migrations.AlterField(
            model_name='writer',
            name='pr_society',
            field=models.CharField(blank=True, choices=[('055', 'SABAM, Belgium'), ('101', 'SOCAN, Canada'), ('088', 'CMRRA, Canada'), ('040', 'KODA, Denmark'), ('089', 'TEOSTO, Finland'), ('058', 'SACEM, France'), ('035', 'GEMA, Germany'), ('074', 'SIAE, Italy'), ('023', 'BUMA, Netherlands'), ('078', 'STEMRA, Netherlands'), ('090', 'TONO, Norway'), ('079', 'STIM, Sweden'), ('052', 'PRS, United Kingdom'), ('044', 'MCPS, United Kingdom'), ('010', 'ASCAP, United States'), ('021', 'BMI, United States'), ('071', 'SESAC Inc., United States'), ('034', 'HFA, United States')], help_text='Required for a controlled writer', max_length=3, null=True, validators=[music_publisher.models.CWRFieldValidator('writer_pr_society')], verbose_name='Performing Rights Society'),
        ),
        migrations.AlterField(
            model_name='writer',
            name='publisher_fee',
            field=models.DecimalField(blank=True, decimal_places=2, help_text='Percentage of royalties kept by the publisher', max_digits=5, null=True, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)]),
        ),
        migrations.AlterField(
            model_name='writerinwork',
            name='publisher_fee',
            field=models.DecimalField(blank=True, decimal_places=2, help_text='Percentage of royalties kept by the publisher', max_digits=5, null=True, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)]),
        ),
        migrations.AlterUniqueTogether(
            name='albumcd',
            unique_together={('album_title', 'album_label')},
        ),
    ]



"""
This app is targeted at single original publishers, publishing original musical
works. It holds data on musical works, including alternate titles, songwriters,
performing artists, first recordings, music library and albums. It allows data
imports and exports.

Multiple writers, both controlled and uncontrolled, are covered, with minor
limitations, but data on other publishers (other original publishers,
administrators and sub-publishers) can not be entered.

A special US situation where an original publisher may have one entity for
every of the three PROs is also covered.
"""



"""Tests for :mod:`music_publisher`.

Tests are the weakest link in the package. They do test that things do not break, which makes a grat difference, but they do not test how things work.

A major re-write is required.

Attributes:
    CONTENT (str): CWR ACK file contents
"""

from datetime import date, time
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.test import TestCase, Client, override_settings
from django.test.client import RequestFactory
from django.urls import reverse
from music_publisher.admin import *
from music_publisher.models import *
from io import StringIO


class AllTest(TestCase):
    """All Tests in a sibgle class.
    """

    def setUp(self):
        """Set up the initial data.
        """
        self.adminsite = AdminSite()
        self.user = User(
            username='superuser', is_superuser=True, is_staff=True)
        self.user.save()
        self.client = Client()
        self.client.force_login(self.user)
        self.factory = RequestFactory()

        self.artist = Artist(last_name="FOO", first_name="")
        self.artist.save()
        self.recordingartist = Artist(last_name="BAR", first_name="BARBARA")
        self.recordingartist.save()
        self.thiswriter = Writer(
            last_name='FOOBAR', first_name='JACK', ipi_name='199',
            ipi_base='I1234567893', pr_society='010')
        self.thiswriter.full_clean()
        self.thiswriter.save()
        self.thatwriter = Writer(
            last_name='FOOBAR', first_name='JACK JR.', ipi_name='297',
            pr_society='010', generally_controlled=True, saan='JaVa',
            publisher_fee=50)
        self.thatwriter.full_clean()
        self.thatwriter.save()
        self.otherwriter = Writer(last_name='DOE', first_name='JANE')
        self.otherwriter.full_clean()
        self.otherwriter.save()
        self.album_cd = AlbumCD(
            album_title='ALBUM', cd_identifier='XXX 123',
            release_date=date.today())
        self.album_cd.save()
        self.work = Work(title='THE TITLE', iswc='T-123.456.789-4')
        self.work.save()
        ArtistInWork(artist=self.artist, work=self.work).save()
        WriterInWork(
            writer=self.thiswriter, work=self.work, saan='SAMOSAN',
            publisher_fee='20',
            controlled=True, relative_share=50, capacity='CA').save()
        WriterInWork(
            writer=self.otherwriter, work=self.work, relative_share=50,
            controlled=False).save()
        self.alternate_title = AlternateTitle(
            work=self.work, title='The Alternate Title')
        self.alternate_title.save()
        self.work.firstrecording = FirstRecording(
            work=self.work, isrc='US-123-18-10000', duration=time(minute=1),
            album_cd=self.album_cd, artist=self.recordingartist,
            release_date=date.today())
        self.work.firstrecording.save()
        self.work.full_clean()
        self.work2 = Work(title='Minimal')
        self.work2.save()
        WriterInWork(
            writer=self.thiswriter, work=self.work2, saan='SAMOSAN',
            controlled=True, relative_share=50, capacity='CA').save()
        WriterInWork(
            work=self.work2, controlled=False, relative_share=50).save()
        self.cwr_export = CWRExport()
        self.cwr_export.save()
        self.cwr_export.works.add(self.work)
        self.cwr_export.works.add(self.work2)
        self.cwr_export2 = CWRExport(nwr_rev='REV')
        self.cwr_export2.save()
        self.cwr_export2.works.add(self.work)

    @override_settings()
    def test_validation(self):
        """Test the field-level validation for the initial models.
        """
        self.assertTrue(VALIDATE)
        self.work.clean_fields()
        get_publisher_dict(None)
        with self.assertRaises(ValidationError) as ve:
            self.artist.first_name = "BAR, JR"
            self.artist.clean_fields()
        self.assertIn('first_name', ve.exception.message_dict)
        self.artist.first_name = "BAR JR"
        self.assertFalse(CWRFieldValidator('x') == CWRFieldValidator('y'))
        with self.assertRaises(ValidationError):
            self.artist.validate_fields({'bad': 'dict'})
        self.work.firstrecording.clean_fields()
        self.assertEqual(self.work.firstrecording.isrc, 'US1231810000')
        self.thiswriter.clean_fields()
        self.assertTrue(self.thiswriter._can_be_controlled)
        self.assertFalse(self.otherwriter._can_be_controlled)
        with self.assertRaises(ValidationError) as ve:
            self.otherwriter.saan = 'S4N1LJ4V4'
            self.otherwriter.full_clean()
        self.assertIn('saan', ve.exception.message_dict)
        with self.assertRaises(ValidationError) as ve:
            self.otherwriter.saan = None
            self.otherwriter.generally_controlled = True
            self.otherwriter.full_clean()
        self.assertIn('generally_controlled', ve.exception.message_dict)
        with self.assertRaises(ValidationError) as ve:
            self.thiswriter.saan = 'S4N1LJ4V4'
            self.thiswriter.full_clean()
        self.assertIn('saan', ve.exception.message_dict)
        with self.assertRaises(ValidationError) as ve:
            self.thiswriter.saan = None
            self.thiswriter.publisher_fee = 100
            self.thiswriter.full_clean()
        self.assertIn('publisher_fee', ve.exception.message_dict)
        self.thiswriter.publisher_fee = None
        with self.assertRaises(ValidationError) as ve:
            self.thiswriter.pr_society = None
            self.thiswriter.full_clean()
        with self.assertRaises(ValidationError) as ve:
            self.thatwriter.saan = None
            self.thatwriter.full_clean()
        self.assertIn('saan', ve.exception.message_dict)
        with self.assertRaises(ValidationError) as ve:
            self.thatwriter.saan = 'SANILJAVA'
            self.thatwriter.publisher_fee = None
            self.thatwriter.full_clean()
        self.assertIn('publisher_fee', ve.exception.message_dict)
        wiws = self.work.writerinwork_set.all()
        for wiw in wiws:
            if not wiw.controlled:
                with self.assertRaises(ValidationError) as ve:
                    wiw.writer.generally_controlled = True
                    wiw.full_clean()
                self.assertIn('controlled', ve.exception.message_dict)
                wiw.writer.generally_controlled = False
                with self.assertRaises(ValidationError) as ve:
                    wiw.publisher_fee = 100
                    wiw.full_clean()
                self.assertIn('publisher_fee', ve.exception.message_dict)
                wiw.publisher_fee = None
                with self.assertRaises(ValidationError) as ve:
                    wiw.saan = 'JAVA'
                    wiw.full_clean()
                self.assertIn('saan', ve.exception.message_dict)
                wiw.saan = None
            else:
                with self.assertRaises(ValidationError) as ve:
                    wiw.capacity = None
                    wiw.full_clean()
                self.assertIn('capacity', ve.exception.message_dict)
                wiw.capacity = 'CA'
                with self.assertRaises(ValidationError) as ve:
                    wiw.writer._can_be_controlled = False
                    wiw.full_clean()
                self.assertIn('__all__', ve.exception.message_dict)
                wiw.writer._can_be_controlled = True
                if wiw.writer and not wiw.writer.generally_controlled:
                    with self.assertRaises(ValidationError) as ve:
                        wiw.saan = None
                        wiw.full_clean()
                    self.assertIn('saan', ve.exception.message_dict)
                    with self.assertRaises(ValidationError) as ve:
                        wiw.saan = 'SIJ'
                        wiw.publisher_fee = None
                        wiw.full_clean()
                    self.assertIn('publisher_fee', ve.exception.message_dict)
                with self.assertRaises(ValidationError) as ve:
                    wiw.writer = None
                    wiw.full_clean()
                self.assertIn('writer', ve.exception.message_dict)
        settings.MUSIC_PUBLISHER_SETTINGS['validator_url'] = \
            'https://240.0.0.1/'  # no routing possible to this address
        self.work.full_clean()
        self.assertFalse(self.work._cwr)
        self.work.full_clean()

    def test_str(self):
        """Test that all strings are in CWR_compliant form.
        """
        self.assertEqual(str(self.artist), 'FOO')
        self.artist.first_name = "BAR JR"
        self.assertEqual(str(self.artist), 'BAR JR FOO')
        self.assertEqual(str(self.alternate_title), 'THE ALTERNATE TITLE')
        self.assertEqual(str(self.album_cd), 'ALBUM (XXX 123)')
        self.assertEqual(self.album_cd.album_label, SETTINGS.get('label'))
        self.assertEqual(self.album_cd.library, SETTINGS.get('library'))
        self.album_cd.album_title = None
        self.assertEqual(str(self.album_cd), 'XXX 123')
        with self.assertRaises(ValidationError) as ve:
            self.album_cd.full_clean()
        self.assertIn('album_title', ve.exception.message_dict)
        self.album_cd.cd_identifier = None
        with self.assertRaises(ValidationError) as ve:
            self.album_cd.full_clean()
        self.assertIn('cd_identifier', ve.exception.message_dict)
        self.assertIn('album_title', ve.exception.message_dict)
        self.assertEqual(str(self.thiswriter), 'JACK FOOBAR')
        self.thiswriter.generally_controlled = True
        self.assertEqual(str(self.thiswriter), 'JACK FOOBAR (*)')
        self.assertEqual(str(self.work.firstrecording), str(self.work))
        aiw = self.work.artistinwork_set.first()
        self.assertEqual(str(aiw), 'FOO')
        wiw = self.work.writerinwork_set.first()
        self.assertEqual(str(wiw), str(wiw.writer))

    @override_settings()
    def test_cwr(self):
        """Test external CWR generation.
        """
        self.assertIn('NWR DRAFT ', self.cwr_export.filename)
        self.cwr_export.get_cwr()
        self.assertIn('CW', self.cwr_export.filename)
        self.cwr_export.get_cwr()
        self.cwr_export2.get_cwr()
        self.cwr_export.cwr = None
        token = settings.MUSIC_PUBLISHER_SETTINGS['token']
        settings.MUSIC_PUBLISHER_SETTINGS['token'] = 'aaa'
        with self.assertRaises(ValidationError) as ve:
            self.cwr_export.get_cwr()
        self.assertIn('Unauthorized', ve.exception.message)
        settings.MUSIC_PUBLISHER_SETTINGS['token'] = token
        url = settings.MUSIC_PUBLISHER_SETTINGS['generator_url']
        settings.MUSIC_PUBLISHER_SETTINGS['generator_url'] = None
        with self.assertRaises(ValidationError) as ve:
            self.cwr_export.get_cwr()
        settings.MUSIC_PUBLISHER_SETTINGS['generator_url'] = \
            'https://240.0.0.1/'  # no routing possiblle
        with self.assertRaises(ValidationError) as ve:
            self.cwr_export.get_cwr()
        settings.MUSIC_PUBLISHER_SETTINGS['generator_url'] = url
        pin = settings.MUSIC_PUBLISHER_SETTINGS['publisher_ipi_name']
        settings.MUSIC_PUBLISHER_SETTINGS['publisher_ipi_name'] = 100
        with self.assertRaises(ValidationError) as ve:
            self.cwr_export.get_cwr()
        settings.MUSIC_PUBLISHER_SETTINGS['publisher_ipi_name'] = pin

    def get(self, path, re_post=False):
        """A helper method that similates opening of view and then simulates
            manual changes and save.
        """
        response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        adminform = response.context_data.get('adminform')
        if adminform is None:
            return
        if not re_post:
            return response
        data = {}
        for sc in response.context:
            for d in sc:
                if 'widget' in sc:
                    data[sc['widget']['name']] = sc['widget']['value']
        data.update(adminform.form.initial)
        if isinstance(re_post, dict):
            data.update(re_post)
        for key, value in data.items():
            if value is None:
                data[key] = ''
            else:
                data[key] = value
        response = self.client.post(path, data)
        self.assertIn(response.status_code, [200, 302])
        return response

    @override_settings()
    def test_admin(self):
        """Integration tests for admin interface.
        """
        self.assertEqual(AlbumCDAdmin.label_not_set(None), 'NOT SET')
        self.cwr_export.get_cwr()
        self.cwr_export2.get_cwr()
        request = self.client.get(
            reverse('admin:music_publisher_ackimport_add',)).wsgi_request
        ACKImportAdmin(
            ACKImport, self.adminsite).process(request, '021', CONTENT)
        self.get(reverse('admin:index'))
        self.get(reverse('admin:app_list', args=('music_publisher',)))
        self.get(reverse('admin:music_publisher_artist_changelist',))
        self.get(reverse('admin:music_publisher_albumcd_changelist',))
        self.get(reverse('admin:music_publisher_writer_changelist',))
        self.get(reverse('admin:music_publisher_work_changelist',))
        self.get(
            reverse('admin:music_publisher_work_changelist',) +
            '?has_iswc=N&has_rec=Y')
        self.get(
            reverse('admin:music_publisher_work_changelist',) +
            '?has_iswc=Y&has_rec=N&q=01')
        self.get(reverse('admin:music_publisher_cwrexport_changelist',))
        self.get(reverse('admin:music_publisher_ackimport_changelist',))
        self.get(reverse('admin:music_publisher_artist_add',) + '?_popup=1')
        self.get(reverse('admin:music_publisher_albumcd_add',) + '?_popup=1')
        self.get(reverse('admin:music_publisher_work_add',) + '?_popup=1')
        self.get(reverse('admin:music_publisher_artist_add',))
        self.get(reverse('admin:music_publisher_albumcd_add',))
        self.get(reverse('admin:music_publisher_writer_add',))
        self.get(reverse('admin:music_publisher_work_add',))
        self.get(reverse('admin:music_publisher_cwrexport_add',))
        self.get(reverse('admin:music_publisher_ackimport_add',))
        self.get(reverse(
            'admin:music_publisher_artist_change', args=(self.artist.id,)),
            re_post={'last_name': 'BAR'})
        self.assertEqual(str(Artist.objects.get(pk=self.artist.id)), 'BAR')
        url = settings.MUSIC_PUBLISHER_SETTINGS['validator_url']
        settings.MUSIC_PUBLISHER_SETTINGS['validator_url'] = \
            'https://240.0.0.1/'  # no routing possiblle
        self.get(reverse(
            'admin:music_publisher_artist_change', args=(self.artist.id,)),
            re_post={'last_name': 'CAFE'})
        settings.MUSIC_PUBLISHER_SETTINGS['validator_url'] = url
        self.get(reverse(
            'admin:music_publisher_albumcd_change', args=(self.album_cd.id,)),
            re_post={'album_title': None, 'release_date': None})
        self.assertEqual(
            str(AlbumCD.objects.get(pk=self.album_cd.id)), 'XXX 123')
        self.get(reverse(
            'admin:music_publisher_writer_change', args=(self.thiswriter.id,)),
            re_post={'pr_society': '021'})
        self.get(reverse(
            'admin:music_publisher_work_change', args=(self.work.id,)),
            re_post={'title': 'CHANGED', 'writerinwork_set-0-controlled': 1})
        self.get(reverse(
            'admin:music_publisher_work_change', args=(self.work.id,)),
            re_post={
                'writerinwork_set-1-relative_share': '111',
                'writerinwork_set-0-controlled': 1})
        self.get(reverse(
            'admin:music_publisher_work_change', args=(self.work.id,)),
            re_post={
                'writerinwork_set-0-relative_share': '50',
                'writerinwork_set-1-relative_share': '50',
                'writerinwork_set-0-controlled': 0,
                'writerinwork_set-1-controlled': 0})
        self.get(reverse(
            'admin:music_publisher_work_change', args=(self.work.id,)),
            re_post={
                'writerinwork_set-1-relative_share': 'BAD DATA'})
        self.get(reverse(
            'admin:music_publisher_albumcd_change', args=(self.album_cd.id,)),
            re_post={'album_title': 'CHANGED', 'library': 'SET'})
        self.client.post(
            reverse('admin:music_publisher_work_changelist'),
            data={
                'action': 'create_cwr', 'select_across': 1,
                'index': 0, '_selected_action': self.work.id})
        self.client.post(
            reverse('admin:music_publisher_work_changelist'),
            data={
                'action': 'create_json', 'select_across': 1,
                'index': 0, '_selected_action': self.work.id})
        self.get(
            reverse(
                'admin:music_publisher_cwrexport_change',
                args=(self.cwr_export.id,)),
            re_post=True)
        self.client.get(reverse(
            'admin:music_publisher_cwrexport_change',
            args=(self.cwr_export.id,)) + '?download=true',)
        self.client.get(reverse(
            'admin:music_publisher_cwrexport_change',
            args=(self.cwr_export.id,)) + '?preview=true',)
        settings.MUSIC_PUBLISHER_SETTINGS['highlighter_url'] = \
            'https://240.0.0.1/'  # no routing possiblle
        self.get(
            reverse('admin:music_publisher_cwrexport_add'),
            re_post={'nwr_rev': 'NWR', 'works': [1]})
        self.get(
            reverse('admin:music_publisher_cwrexport_add'),
            re_post={'nwr_rev': 'WRK', 'works': [1]})
        for export in CWRExport.objects.order_by('-id')[0:2]:
            self.client.get(reverse(
                'admin:music_publisher_cwrexport_change',
                args=(export.id,)) + '?preview=true',)
        mock = StringIO()
        mock.write(CONTENT)
        mock.seek(0)
        mockfile = InMemoryUploadedFile(
            mock, 'acknowledgement_file', 'CW180001000_FOO.V21',
            'text', 0, None)
        self.get(
            reverse('admin:music_publisher_ackimport_add'),
            re_post={'acknowledgement_file': mockfile})
        mock.seek(3)
        mockfile = InMemoryUploadedFile(
            mock, 'acknowledgement_file', 'CW180001000_FOO.V21',
            'text', 0, None)
        self.get(
            reverse('admin:music_publisher_ackimport_add'),
            re_post={'acknowledgement_file': mockfile})
        mock.seek(3)
        mockfile = InMemoryUploadedFile(
            mock, 'acknowledgement_file', 'CW180001000_FOO.V21',
            'text', 0, None)
        self.get(
            reverse('admin:music_publisher_ackimport_add'),
            re_post={'acknowledgement_file': mockfile})
        mock.seek(0)
        mockfile = InMemoryUploadedFile(
            mock, 'acknowledgement_file', 'CW180001000_FOOL.V21',
            'text', 0, None)
        self.get(
            reverse('admin:music_publisher_ackimport_add'),
            re_post={'acknowledgement_file': mockfile})
        self.get(
            reverse(
                'admin:music_publisher_ackimport_change',
                args=(1,)),
            re_post=True)


CONTENT = """HDRSO000000021BMI                                          01.102018060715153220180607
GRHACK0000102.100020180607
ACK0000000000000000201805160910510000100000000NWRONE                                                         00000000000001                          20180607AS
ACK0000000100000000201805160910510000100000001NWRTWO                                                         00000000000002                          20180607AS
ACK0000000200000000201805160910510000100000002NWRTHREE                                                       00000000000003                          20180607NP
ACK0000000300000000201805160910510000100000003NWRX                                                           0000000000000X                          20180607NP
TRL000010000008000000839"""


# Generated by Django 2.1.3 on 2018-11-14 15:48

from django.db import migrations, models
import music_publisher.base


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0009_auto_20180912_1217'),
    ]

    operations = [
        migrations.AddField(
            model_name='artist',
            name='isni',
            field=models.CharField(blank=True, null=True, max_length=16, validators=[music_publisher.base.CWRFieldValidator('isni')]),
        ),
        migrations.AlterField(
            model_name='workacknowledgement',
            name='society_code',
            field=models.CharField(choices=[('055', 'SABAM, Belgium'), ('101', 'SOCAN, Canada'), ('088', 'CMRRA, Canada'), ('040', 'KODA, Denmark'), ('089', 'TEOSTO, Finland'), ('058', 'SACEM, France'), ('035', 'GEMA, Germany'), ('074', 'SIAE, Italy'), ('023', 'BUMA, Netherlands'), ('078', 'STEMRA, Netherlands'), ('090', 'TONO, Norway'), ('079', 'STIM, Sweden'), ('052', 'PRS, United Kingdom'), ('044', 'MCPS, United Kingdom'), ('010', 'ASCAP, United States'), ('021', 'BMI, United States'), ('071', 'SESAC Inc., United States'), ('034', 'HFA, United States'), ('319', 'ICE Services, Administrative Agency'), ('707', 'Musicmark, Administrative Agency')], max_length=3, verbose_name='Society'),
        ),
        migrations.AlterField(
            model_name='workacknowledgement',
            name='status',
            field=models.CharField(choices=[('CO', 'Conflict'), ('DU', 'Duplicate'), ('RA', 'Transaction Accepted'), ('AS', 'Registration Accepted'), ('AC', 'Registration Accepted with Changes'), ('RJ', 'Rejected'), ('NP', 'No Participation'), ('RC', 'Claim rejected')], max_length=2),
        ),
        migrations.AlterField(
            model_name='writer',
            name='ipi_name',
            field=models.CharField(blank=True, max_length=11, null=True, unique=True, validators=[music_publisher.base.CWRFieldValidator('writer_ipi_name')], verbose_name='IPI Name #'),
        ),
        migrations.AlterField(
            model_name='writer',
            name='pr_society',
            field=models.CharField(blank=True, choices=[('055', 'SABAM, Belgium'), ('101', 'SOCAN, Canada'), ('088', 'CMRRA, Canada'), ('040', 'KODA, Denmark'), ('089', 'TEOSTO, Finland'), ('058', 'SACEM, France'), ('035', 'GEMA, Germany'), ('074', 'SIAE, Italy'), ('023', 'BUMA, Netherlands'), ('078', 'STEMRA, Netherlands'), ('090', 'TONO, Norway'), ('079', 'STIM, Sweden'), ('052', 'PRS, United Kingdom'), ('044', 'MCPS, United Kingdom'), ('010', 'ASCAP, United States'), ('021', 'BMI, United States'), ('071', 'SESAC Inc., United States'), ('034', 'HFA, United States'), ('319', 'ICE Services, Administrative Agency'), ('707', 'Musicmark, Administrative Agency')], max_length=3, null=True, validators=[music_publisher.base.CWRFieldValidator('writer_pr_society')], verbose_name='Performing Rights Society'),
        ),
    ]



"""Contains base (abstract) classes used in :mod:`.models`

Attributes:
    CAN_NOT_BE_CONTROLLED_MSG (str): \
        Error message when required fields are missing
    CONTROLLED_WRITER_REQUIRED_FIELDS (list): List of required fields
    ENFORCE_IPI_NAME (bool): \
        Is IPI Name # a required field for controlled writers?
    ENFORCE_PR_SOCIETY (TYPE): \
        Is PR Society Code a required field for controlled writers?
    ENFORCE_PUBLISHER_FEE (TYPE): \
        Is Publisher Fee a required field for controlled writers?
    ENFORCE_SAAN (TYPE): \
        Is Society-assigned agreement number a required field for controlled \
        writers?
    VALIDATE (TYPE): Are settings for remote validation service present?
    WORK_ID_PREFIX (TYPE): Prefix for Submitter Work ID, which is numerical
"""

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.deconstruct import deconstructible
import re
import requests


# SETTINGS, strictly not required, except for validation and CWR generation

if hasattr(settings, 'MUSIC_PUBLISHER_SETTINGS'):
    SETTINGS = settings.MUSIC_PUBLISHER_SETTINGS
else:
    SETTINGS = {}

ENFORCE_SAAN = SETTINGS.get('enforce_saan')
ENFORCE_PUBLISHER_FEE = SETTINGS.get('enforce_publisher_fee')
ENFORCE_PR_SOCIETY = SETTINGS.get('enforce_pr_society')
ENFORCE_IPI_NAME = SETTINGS.get('enforce_ipi_name')
CONTROLLED_WRITER_REQUIRED_FIELDS = ['Last Name']
if ENFORCE_PR_SOCIETY:
    CONTROLLED_WRITER_REQUIRED_FIELDS.append('Performing Rights Society')
if ENFORCE_IPI_NAME:
    CONTROLLED_WRITER_REQUIRED_FIELDS.append('IPI Name #')
CAN_NOT_BE_CONTROLLED_MSG = (
    'Unsufficient data for a controlled writer, required fields are: {}.'
).format(', '.join(CONTROLLED_WRITER_REQUIRED_FIELDS))
WORK_ID_PREFIX = SETTINGS.get('work_id_prefix') or ''
VALIDATE = (
    SETTINGS.get('validator_url') and
    SETTINGS.get('highlighter_url') and
    SETTINGS.get('generator_url') and
    SETTINGS.get('token'))


# Default only has 18 societies from 12 countries. These are G7, without Japan,
# where we have no information about CWR use and all other societies covered
# by ICE Services
# If you need to add some, do it in the settings, not here.

try:
    SOCIETIES = settings.MUSIC_PUBLISHER_SOCIETIES
except AttributeError:
    SOCIETIES = [
        ('055', 'SABAM, Belgium'),

        ('101', 'SOCAN, Canada'),
        ('088', 'CMRRA, Canada'),

        ('040', 'KODA, Denmark'),

        ('089', 'TEOSTO, Finland'),

        ('058', 'SACEM, France'),

        ('035', 'GEMA, Germany'),

        ('074', 'SIAE, Italy'),

        ('023', 'BUMA, Netherlands'),
        ('078', 'STEMRA, Netherlands'),

        ('090', 'TONO, Norway'),

        ('079', 'STIM, Sweden'),

        ('052', 'PRS, United Kingdom'),
        ('044', 'MCPS, United Kingdom'),

        ('010', 'ASCAP, United States'),
        ('021', 'BMI, United States'),
        ('071', 'SESAC Inc., United States'),
        ('034', 'HFA, United States'),

        ('319', 'ICE Services, Administrative Agency'),
        ('707', 'Musicmark, Administrative Agency')]


def get_publisher_dict(pr_society):
    """Return publisher settings based on society code.

    Args:
        pr_society (str): CISAC Society code, 3-digit format with leading zero.

    Returns:
        dict: Data on publisher in question, default one if not found.
    """

    mapping = {'010': 'ASCAP', '021': 'BMI', '071': 'SESAC'}
    key = mapping.get(pr_society)
    us_publisher_override = SETTINGS.get('us_publisher_override', {})
    pub = us_publisher_override.get(key)
    if pub:
        pub['pr_society'] = 'pr_society'
        return pub
    return SETTINGS


@deconstructible
class CWRFieldValidator:
    """Validate fields for CWR compliance.

    This validator that does not really validate, it just sets the correct
    field type.
    Fields are validated in batches in :meth:`MusicPublisherBase.clean_fields`.

    Attributes:
        field (str): Validation service name of the field being validated
    """

    field = ''

    def __init__(self, field):
        """Initialize the validator.

        Args:
            field (str): Validation service name of the field being validated
        """
        self.field = field

    def __call__(self, value):
        """Dummy call

        Args:
            value (): Not used, dummy call

        Returns:
            None: dummy call
        """
        return


class MusicPublisherBase(models.Model):
    """Abstract class for all top-level classes.
    """

    class Meta:
        abstract = True

    _cwr = models.BooleanField(
        'CWR-Ready', editable=False, default=False)

    def validate_fields(self, fields):
        """Validate the fields with an external service.

        Args:
            fields (dict): dict of dicts, see :meth:`construct_field_dict`

        Raises:
            ValidationError: If remote validation is not successful
        """

        keys = list(fields.keys())
        values = list(fields.values())
        data = {'fields': values}
        try:
            response = requests.post(
                SETTINGS.get('validator_url'),
                headers={'Authorization': 'Token {}'.format(
                    SETTINGS.get('token'))},
                json=data, timeout=10)
        except requests.exceptions.ConnectionError:
            raise ValidationError('Network error', code='service')
        if response.status_code != 200:
            raise ValidationError('Validation failed', code='service')
        errors = {}
        rfields = response.json()['fields']
        for i in range(len(fields)):
            field_name = keys[i]
            field_dict = rfields[i]
            if not field_dict.get('conditionally_valid'):
                errors[field_name] = (
                    field_dict['error'] or 'Unknown Error')
        if errors:
            raise ValidationError(errors, code='invalid')
        self._cwr = True

    def construct_field_dict(self):
        """Construct dictionary appropriate for external validation.

        Returns:
            dict: Dictionary with following structure:
                {local_field_name: {
                    'field': validator_field_name,
                    'value': field-value}}
        """
        fields = {}
        for field in self._meta.fields:
            if not hasattr(self, field.name):
                continue
            value = getattr(self, field.name)
            if not value:
                continue
            for validator in field.validators:
                if isinstance(validator, CWRFieldValidator):
                    fields[field.name] = {
                        'field': validator.field,
                        'value': value}
        return fields

    def clean_fields(self, *args, **kwargs):
        """If external service is used, prepare the data for validation,
        and perform the validation.
        """

        if not VALIDATE:  # external validation not active, just skip
            self._cwr = False
            return super().clean_fields(*args, **kwargs)

        fields = self.construct_field_dict()
        try:
            self.validate_fields(fields)
        except ValidationError as ve:
            self._cwr = False
            if 'service' not in ve.args:
                # This means it is an unknown error
                raise
        return super().clean_fields(*args, **kwargs)


class TitleBase(MusicPublisherBase):
    """Abstract class for all classes that have a title.

    Attributes:
        title (django.db.models.CharField): Title, used in work title,
            alternate title, etc.
    """

    class Meta:
        abstract = True
        ordering = ('-id',)

    title = models.CharField(
        max_length=60, db_index=True, validators=(
            # Original CWR Generator marks all alternate titles as 'AT',
            # so the validation is the same as in NWR.work_title
            CWRFieldValidator('work_title'),))

    def __str__(self):
        return self.title.upper()


class WorkBase(TitleBase):
    """Abstract class for musical works, most important top-level class.

    Attributes:
        iswc (django.db.models.CharField):
            ISWC - International Standard Work Code
    """

    class Meta:
        abstract = True
        verbose_name_plural = '  Works'

    iswc = models.CharField(
        'ISWC', max_length=15, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('iswc'),))

    def clean_fields(self, *args, **kwargs):
        """Deal with various ways ISWC is written.
        """
        if self.iswc:
            # CWR 2.x holds ISWC in TNNNNNNNNNC format
            # CWR 3.0 holds ISWC in T-NNNNNNNNN-C format
            # sometimes it comes in T-NNN.NNN.NNN-C format
            self.iswc = self.iswc.replace('-', '').replace('.', '')
        return super().clean_fields(*args, **kwargs)


class AlbumCDBase(MusicPublisherBase):
    """Abstract class that deals with albums and music liibrary data.

    Note that label and library are set in the settings, serving as defualts.

    Attributes:
        album_label (django.db.models.CharField): Album Label
        album_title (django.db.models.CharField): Album Title
        cd_identifier (django.db.models.CharField):
            CD Indetifier (for Library works)
        ean (django.db.models.CharField): Album (or other release) EAN
        release_date (django.db.models.DateField):
            The date of the album release, can be overridden in recordings
    """

    class Meta:
        abstract = True
        verbose_name = 'Album and/or Library CD'
        verbose_name_plural = ' Albums and Library CDs'
        ordering = ('album_title', 'cd_identifier')

    cd_identifier = models.CharField(
        'CD Identifier',
        help_text='This will set the purpose to Library.',
        max_length=15, blank=True, null=True, unique=True, validators=(
            CWRFieldValidator('cd_identifier'),))
    release_date = models.DateField(
        help_text='Can be overridden by recording data.',
        blank=True, null=True)
    album_title = models.CharField(
        max_length=60, blank=True, null=True, unique=True, validators=(
            CWRFieldValidator('first_album_title'),))
    ean = models.CharField(
        'EAN',
        max_length=13, blank=True, null=True, unique=True, validators=(
            CWRFieldValidator('ean'),))
    album_label = models.CharField(
        default=SETTINGS.get('label') or '',
        max_length=60, blank=True, validators=(
            CWRFieldValidator('first_album_label'),))

    def __str__(self):
        if self.cd_identifier and self.album_title:
            return '{} ({})'.format(
                self.album_title or '', self.cd_identifier).upper()
        return (self.album_title or self.cd_identifier).upper()

    @property
    def library(self):
        """Return Library name if cd_identifier is present

        Returns:
            str: Name of the library, or None
        """
        if self.cd_identifier:
            return SETTINGS.get('library')

    def clean(self):
        if not self.cd_identifier and not self.album_title:
            raise ValidationError({
                'cd_identifier': 'Required if Album Title is not set.',
                'album_title': 'Required if CD Identifier is not set.'})
        if (self.ean or self.release_date) and not self.album_title:
            raise ValidationError({
                'album_title': 'Required if EAN or release date is set.'})


class RecordingBase(MusicPublisherBase):
    """Holds data on first recording.

    Note that the CWR 2.x limitation of just one REC record per work has been
    removed in the specs, but some societies still complain about it,
    so only a single instance is allowed.

    Attributes:
        duration (django.db.models.TimeField): Recording Duration
        isrc (django.db.models.CharField):
            International Standard Recording Code
        record_label (django.db.models.CharField): Record Label
        release_date (django.db.models.DateField): Recording Release Date
    """

    class Meta:
        abstract = True
        verbose_name_plural = 'First recording of the work'

    release_date = models.DateField(blank=True, null=True)
    duration = models.TimeField(blank=True, null=True)
    isrc = models.CharField(
        'ISRC', max_length=15, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('isrc'),))
    record_label = models.CharField(
        default=SETTINGS.get('label') or '',
        max_length=60, blank=True, validators=(
            CWRFieldValidator('first_album_label'),))

    def clean_fields(self, *args, **kwargs):
        if self.isrc:
            # Removing all characters added for readability
            self.isrc = self.isrc.replace('-', '').replace('.', '')
        return super().clean_fields(*args, **kwargs)


class PersonBase(models.Model):
    """Base class for all classes that contain people with first and last name.

    Attributes:
        first_name (django.db.models.CharField): First Name
        last_name (django.db.models.CharField): Last Name
    """

    class Meta:
        abstract = True

    first_name = models.CharField(
        max_length=30, blank=True,
        validators=(CWRFieldValidator('artist_first_name'),))
    last_name = models.CharField(
        max_length=45, db_index=True,
        validators=(CWRFieldValidator('artist_last_name'),))

    def __str__(self):
        if self.first_name:
            return '{0.first_name} {0.last_name}'.format(self).upper()
        return self.last_name.upper()


class IPIBase(models.Model):
    """Abstract base for all objects containing IPI numbers.

    Attributes:
        generally_controlled (django.db.models.BooleanField):
            General agreement (renamed in verbose_name)
        ipi_base (django.db.models.CharField): IPI Base Number
        ipi_name (django.db.models.CharField): IPI Name Number
        pr_society (django.db.models.CharField):
            Performing Rights Society Code
        publisher_fee (django.db.models.DecimalField): Publisher Fee
        saan (django.db.models.CharField):
            Society-assigned agreement number, in this context it is used for
            general agreements, for specific agreements use
            :attr:`.models.WriterInWork.saan`.
    """

    class Meta:
        abstract = True

    ipi_name = models.CharField(
        'IPI Name #', max_length=11, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('writer_ipi_name'),))
    ipi_base = models.CharField(
        'IPI Base #', max_length=15, blank=True, null=True,
        validators=(CWRFieldValidator('writer_ipi_base'),))
    pr_society = models.CharField(
        'Performing Rights Society', max_length=3, blank=True, null=True,
        validators=(CWRFieldValidator('writer_pr_society'),),
        choices=SOCIETIES)
    saan = models.CharField(
        'Society-assigned agreement number',
        help_text='Use this field for general agreements only.',
        validators=(CWRFieldValidator('saan'),),
        max_length=14, blank=True, null=True, unique=True)

    _can_be_controlled = models.BooleanField(editable=False, default=False)
    generally_controlled = models.BooleanField(
        'General Agreement', default=False)
    publisher_fee = models.DecimalField(
        max_digits=5, decimal_places=2, blank=True, null=True,
        validators=[MinValueValidator(0), MaxValueValidator(100)],
        help_text='Percentage of royalties kept by the publisher')

    def clean_fields(self, *args, **kwargs):
        if self.saan:
            self.saan = self.saan.upper()
        if self.ipi_name:
            self.ipi_name = self.ipi_name.rjust(11, '0')
        if self.ipi_base:
            self.ipi_base = self.ipi_base.replace('.', '')
            self.ipi_base = re.sub(
                r'(I).?(\d{9}).?(\d)', r'\1-\2-\3', self.ipi_base)
        return super().clean_fields(*args, **kwargs)

    def clean(
            self,
            enforce_ipi_name=ENFORCE_IPI_NAME,
            enforce_pr_society=ENFORCE_PR_SOCIETY,
            enforce_saan=ENFORCE_SAAN,
            enforce_publisher_fee=ENFORCE_PUBLISHER_FEE,
            error_msg=CAN_NOT_BE_CONTROLLED_MSG):
        """Clean with a lot of arguments.

        In DMP they come from settings, but in other (closed source) solutions
        that use this code, these values are set dynamically.


        Args:
            enforce_ipi_name (bool, optional):
                Makes IPI Name # required if controlled
            enforce_pr_society (bool, optional):
                Makes PR Society code required if controlled
            enforce_saan (bool, optional):
                Makes SAAN required if controlled
            enforce_publisher_fee (bool, optional):
                Makes Publisher fee required if controlled
            error_msg (str, optional):
                Error Message to show if required fields are not filled out
        """

        self._can_be_controlled = True
        if enforce_ipi_name:
            self._can_be_controlled &= bool(self.ipi_name)
        if enforce_pr_society:
            self._can_be_controlled &= bool(self.pr_society)
        if not self._can_be_controlled and self.generally_controlled:
            raise ValidationError({
                'generally_controlled': error_msg})
        if self.saan and not self.generally_controlled:
            raise ValidationError({
                'saan': 'Only for a general agreement.'})
        if self.publisher_fee and not self.generally_controlled:
            raise ValidationError({
                'publisher_fee': 'Only for a general agreement.'})
        if self.generally_controlled and enforce_saan and not self.saan:
            raise ValidationError({
                'saan': 'This field is required.'})
        if (self.generally_controlled and enforce_publisher_fee and
                not self.publisher_fee):
            raise ValidationError({
                'publisher_fee': 'This field is required.'})


class ArtistBase(PersonBase, MusicPublisherBase):
    """Concrete class for performing artists.

    Attributes:
        isni (django.db.models.CharField):
            International Standard Name Identifier
    """
    class Meta:
        abstract = True
        verbose_name = 'Performing Artist'
        verbose_name_plural = ' Performing Artists'

    isni = models.CharField(
        'ISNI',
        max_length=16, blank=True, null=True, unique=True,
        validators=(CWRFieldValidator('isni'),))


class WriterBase(PersonBase, IPIBase, MusicPublisherBase):
    """Base class for writers, the second most important top-level class.
    """

    class Meta:
        abstract = True
        ordering = ('last_name', 'first_name', 'ipi_name')
        verbose_name_plural = ' Writers'

    _cwr = models.BooleanField(
        'CWR-Compliance', editable=False, default=False)

    def get_publisher_dict(self):
        """Return data on publisher.

        This is the default publisher, except in cases when the writer
        is affiliated with one of the US PROs and this publisher's data
        is set in 'us_publisher_override' of the MUSIC_PUBLISHER_SETTINGS.

        Returns:
            dict: Data on publisher
        """
        return get_publisher_dict(self.pr_society)

    def __str__(self):
        name = super().__str__()
        if self.generally_controlled:
            return name + ' (*)'
        return name



# Generated by Django 2.1.3 on 2018-11-18 18:54

from django.db import migrations, models
import music_publisher.base


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0013_cwrexport__work_count'),
    ]

    operations = [
        migrations.AlterField(
            model_name='albumcd',
            name='album_title',
            field=models.CharField(blank=True, max_length=60, null=True, unique=True, validators=[music_publisher.base.CWRFieldValidator('first_album_title')]),
        ),
        migrations.AlterField(
            model_name='cwrexport',
            name='works',
            field=models.ManyToManyField(related_name='cwr_exports', to='music_publisher.Work'),
        ),
    ]



# Generated by Django 2.1.3 on 2018-12-01 13:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0014_auto_20181118_1854'),
    ]

    operations = [
        migrations.AlterField(
            model_name='albumcd',
            name='_cwr',
            field=models.BooleanField(default=False, editable=False, verbose_name='CWR-Ready'),
        ),
        migrations.AlterField(
            model_name='alternatetitle',
            name='_cwr',
            field=models.BooleanField(default=False, editable=False, verbose_name='CWR-Ready'),
        ),
        migrations.AlterField(
            model_name='artist',
            name='_cwr',
            field=models.BooleanField(default=False, editable=False, verbose_name='CWR-Ready'),
        ),
        migrations.AlterField(
            model_name='firstrecording',
            name='_cwr',
            field=models.BooleanField(default=False, editable=False, verbose_name='CWR-Ready'),
        ),
        migrations.AlterField(
            model_name='work',
            name='_cwr',
            field=models.BooleanField(default=False, editable=False, verbose_name='CWR-Ready'),
        ),
        migrations.AlterUniqueTogether(
            name='albumcd',
            unique_together=set(),
        ),
    ]



from collections import OrderedDict
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
import requests
from .base import *


class Work(WorkBase):
    """Concrete class, with references to foreign objects.

    Attributes:
        artists (django.db.models.ManyToManyField):
            Artists performing the work
        last_change (django.db.models.DateTimeField):
            when the last change was made to this object or any of the child
            objects, basically used in filtering
        writers (django.db.models.ManyToManyField):
            Writers who created the work
    """

    artists = models.ManyToManyField('Artist', through='ArtistInWork')
    writers = models.ManyToManyField('Writer', through='WriterInWork')
    last_change = models.DateTimeField(
        'Last Edited', editable=False, null=True)

    @property
    def json(self):
        """Create a data structure that can be serielized as JSON.

        Returns:
            dict: JSON-serializable data structure
        """

        data = OrderedDict()
        data['work_title'] = self.title
        if self.iswc:
            data['iswc'] = self.iswc
        data['writers'] = [wiw.json for wiw in self.writerinwork_set.all()]
        data['alternate_titles'] = [
            at.json for at in self.alternatetitle_set.all()]
        data['artists'] = [
            aiw.artist.json for aiw in self.artistinwork_set.all()]
        data['society_work_codes'] = [
            wa for wa in self.workacknowledgement_set.filter(
                remote_work_id__gt='').values(
                'society_code', 'remote_work_id').distinct()]
        try:
            data.update(self.firstrecording.json)
        except ObjectDoesNotExist:
            pass
        return {self.work_id: data}

    @property
    def work_id(self):
        """Create Work ID used in registrations

        Returns:
            str: Internal Work ID
        """
        if self.id is None:
            return ''
        return '{}{:06}'.format(WORK_ID_PREFIX, self.id)

    def __str__(self):
        return '{} {} ({})'.format(
            self.work_id,
            self.title,
            ' / '.join(w.last_name.upper() for w in self.writers.all()))


class AlternateTitle(TitleBase):
    """Conrete class for alternate titles.

    Attributes:
        work (django.db.models.ForeignKey): Foreign key to Work model
    """

    work = models.ForeignKey(Work, on_delete=models.CASCADE)

    class Meta:
        unique_together = (('work', 'title'),)

    @property
    def json(self):
        """Create a data structure that can be serielized as JSON.

        Returns:
            dict: JSON-serializable data structure
        """
        return {'alternate_title': self.title}


class AlbumCD(AlbumCDBase):
    """Conrete class for album / CD."""
    pass


class Artist(ArtistBase):
    """Concrete class for performing artists."""

    @property
    def json(self):
        """Create a data structure that can be serielized as JSON.

        Returns:
            dict: JSON-serializable data structure
        """
        return {
            'artist_first_name': self.first_name,
            'artist_last_name': self.last_name,
            'isni': self.isni}


class FirstRecording(RecordingBase):
    """Concrete class for first recording.

    Attributes:
        album_cd (django.db.models.ForeignKey): FK to Album/CD
        artist (django.db.models.ForeignKey):
            FK to Artist, in this context, it is a recording artist
        work (django.db.models.OneToOneField): One to One field to Work
    """

    work = models.OneToOneField(Work, on_delete=models.CASCADE)
    artist = models.ForeignKey(
        Artist, on_delete=models.PROTECT, blank=True, null=True,
        verbose_name='Recording Artist')
    album_cd = models.ForeignKey(
        AlbumCD, on_delete=models.PROTECT, blank=True, null=True,
        verbose_name='Album / Library CD')

    def __str__(self):
        return str(self.work)

    @property
    def json(self):
        """Create a data structure that can be serielized as JSON.

        Returns:
            dict: JSON-serializable data structure
        """
        data = OrderedDict()
        if self.duration:
            data['first_release_duration'] = self.duration.strftime('%H%M%S')
        data['isrc'] = self.isrc or ''
        data['record_label'] = self.record_label
        if self.artist:
            data['recording_artist'] = self.artist.json
        if self.album_cd:
            if self.album_cd.release_date:
                data['first_release_date'] = (
                    self.album_cd.release_date.strftime('%Y%m%d'))
            data.update({
                'first_album_title': self.album_cd.album_title or '',
                'first_album_label': self.album_cd.album_label or '',
                'ean': self.album_cd.ean or ''})
        if self.release_date:
            data['first_release_date'] = self.release_date.strftime('%Y%m%d')
        if self.album_cd and self.album_cd.library:
            data.update({
                'library': self.album_cd.library,
                'cd_identifier': self.album_cd.cd_identifier})
        return data


class ArtistInWork(models.Model):
    """Artist performing the work (live in CWR 3).

    Attributes:
        artist (django.db.models.ForeignKey): FK to Artist
        work (django.db.models.ForeignKey): FK to Work
    """

    work = models.ForeignKey(Work, on_delete=models.CASCADE)
    artist = models.ForeignKey(Artist, on_delete=models.PROTECT)

    class Meta:
        verbose_name = 'Artist performing work'
        verbose_name_plural = 'Artists performing works'
        unique_together = (('work', 'artist'),)

    def __str__(self):
        return str(self.artist)


class Writer(WriterBase):
    """Concrete class for writers.
    """

    def clean(self, *args, **kwargs):
        super().clean(*args, **kwargs)
        if self.pk is None or self._can_be_controlled:
            return
        # A controlled writer requires more data, so once a writer is in
        # that role, it is not allowed to remove required data."""
        if self.writerinwork_set.filter(controlled=True).exists():
            raise ValidationError(
                'This writer is controlled in at least one work. ' +
                CAN_NOT_BE_CONTROLLED_MSG)


class WriterInWork(models.Model):
    """Writers who created this work.

    At least one writer in work must be controlled.
    Sum of relative shares must be (roughly) 100%.
    Capacity is limited to roles for original writers.

    Attributes:
        capacity (django.db.models.CharField): Role of the writer in this work
        controlled (django.db.models.BooleanField): A complete mistery field
        publisher_fee (django.db.models.DecimalField): Percentage of royalties
            kept by publisher
        relative_share (django.db.models.DecimalField): Initial split among
            writers, prior to publishing
        saan (django.db.models.CharField): Society-assigned agreement number
            between the writer and the original publisher, please note that
            this field is for SPECIFIC agreements, for a general agreement,
            use :attr:`.base.IPIBase.saan`
        work (django.db.models.ForeignKey): FK to Work
        writer (django.db.models.ForeignKey): FK to Writer
    """

    class Meta:
        verbose_name_plural = 'Writers in Work'
        unique_together = (('work', 'writer'),)
        ordering = ('-controlled', 'writer__last_name', 'writer__first_name')

    work = models.ForeignKey(Work, on_delete=models.CASCADE)
    writer = models.ForeignKey(
        Writer, on_delete=models.PROTECT,
        blank=True, null=True)
    saan = models.CharField(
        'Society-assigned agreement number',
        help_text='Use this field for specific agreements only.',
        max_length=14, blank=True, null=True,
        validators=(CWRFieldValidator('saan'),),)
    controlled = models.BooleanField(default=False)
    relative_share = models.DecimalField(max_digits=5, decimal_places=2)
    capacity = models.CharField(max_length=2, blank=True, choices=(
        ('C ', 'Composer'),
        ('A ', 'Lyricist'),
        ('CA', 'Composer&Lyricist'),
    ))
    publisher_fee = models.DecimalField(
        max_digits=5, decimal_places=2, blank=True, null=True,
        validators=[MinValueValidator(0), MaxValueValidator(100)],
        help_text='Percentage of royalties kept by the publisher')

    def __str__(self):
        return str(self.writer)

    def clean_fields(self, *args, **kwargs):
        if self.saan:
            self.saan = self.saan.upper()
        return super().clean_fields(*args, **kwargs)

    def clean(self):
        """Make sure that controlled writers have all the required data.

        Also check that writers that are not controlled do not have data
        that can not apply to them."""

        if (self.writer and self.writer.generally_controlled and
                not self.controlled):
            raise ValidationError({
                'controlled': 'Must be set for a generally controlled writer.'
            })
        if self.controlled and not self.capacity:
            raise ValidationError({
                'capacity': 'Must be set for a controlled writer.'
            })
        if self.controlled and not self.writer:
            raise ValidationError({
                'writer': 'Must be set for a controlled writer.'
            })
        if (self.controlled and not self.writer._can_be_controlled):
            raise ValidationError(
                'Writer \'{}\' is controlled in at least one work, '
                'required are: Last name, IPI Name # and Performing '
                'Rights Society.'.format(self.writer))
        if (ENFORCE_SAAN and self.controlled and
                not self.writer.generally_controlled and not self.saan):
            raise ValidationError({
                'saan': 'Must be set. (controlled, no general agreement)'})
        if (ENFORCE_PUBLISHER_FEE and self.controlled and
                not self.writer.generally_controlled and
                not self.publisher_fee):
            raise ValidationError({
                'publisher_fee': (
                    'Must be set. (controlled, no general agreement)')})
        if not self.controlled and self.saan:
            raise ValidationError({'saan': 'Must not be set.'})
        if not self.controlled and self.publisher_fee:
            raise ValidationError({'publisher_fee': 'Must not be set.'})

    @property
    def json(self):
        """Create a data structure that can be serielized as JSON.

        Returns:
            dict: JSON-serializable data structure
        """
        data = OrderedDict()
        if self.writer:
            data['writer_id'] = 'W{:06d}'.format(self.writer.id)
            data['writer_last_name'] = self.writer.last_name or ''
            data['writer_first_name'] = self.writer.first_name or ''
            data['writer_ipi_name'] = self.writer.ipi_name or ''
            data['writer_ipi_base'] = self.writer.ipi_base or ''
            data['writer_pr_society'] = self.writer.pr_society or ''
        else:
            data['writer_id'] = ''
            data['writer_last_name'] = ''
        data.update({
            'controlled': self.controlled,
            'capacity': self.capacity,
            'relative_share': str(self.relative_share / 100),
        })
        if self.controlled:
            publisher = self.writer.get_publisher_dict()
            saan = (
                self.saan or
                (self.writer.saan if self.writer else None) or
                ''
            )
            data.update({
                'publisher_for_writer_id': publisher.get('publisher_id'),
                'saan': saan,
                'general_agreement': bool(saan and not self.saan),
            })
        return data


class CWRExport(models.Model):
    """Export in CWR format.

    Common Works Registration format is a standard format for registration of
    musical works world-wide. As of November 2018, version 2.1r7 is used
    everywhere, while some societies accept 2.2 as well, it adds no benefits
    in this context. Version 3.0 is in draft.

    Attributes:
        nwr_rev (django.db.models.CharField): Choice field where user can
            select which version and type of CWR it is.
    """

    class Meta:
        verbose_name = 'CWR Export'
        verbose_name_plural = 'CWR Exports'
        ordering = ('-id',)

    nwr_rev = models.CharField(
        'NWR/REV', max_length=3, db_index=True, default='NWR', choices=(
            ('NWR', 'CWR 2.1: New work registration'),
            ('REV', 'CWR 2.1: Revision of registered work'),
            ('WRK', 'CWR 3.0: Work registration (alpha)')))
    cwr = models.TextField(blank=True, editable=False)
    year = models.CharField(
        max_length=2, db_index=True, editable=False, blank=True)
    num_in_year = models.PositiveSmallIntegerField(default=0)
    works = models.ManyToManyField(Work, related_name='cwr_exports')
    _work_count = models.IntegerField(
        editable=False, null=True)

    @property
    def is_version_3(self):
        return self.nwr_rev == 'WRK'

    @property
    def filename(self):
        if not self.cwr:
            return '{} DRAFT {}'.format(self.nwr_rev, self.id)
        if self.is_version_3:
            return 'CW{}{:04}{}_000_V3-0-0.SUB'.format(
                self.year,
                self.num_in_year,
                SETTINGS.get('publisher_id'))
        return 'CW{}{:04}{}_000.V21'.format(
            self.year,
            self.num_in_year,
            SETTINGS.get('publisher_id'))

    def __str__(self):
        return self.filename

    def get_json(self, qs=None):
        """Create a data structure that can be serielized as JSON.

        Returns:
            dict: JSON-serializable data structure
        """
        if qs is None:
            qs = self.works.order_by('id')
        if self.is_version_3:
            j = OrderedDict([
                ('filename', self.filename),
                ('version', '3.0'),
            ])
        else:
            j = OrderedDict([('revision', self.nwr_rev == 'REV')])
        j.update({
            "sender_id": SETTINGS.get('publisher_ipi_name'),
            "sender_name": SETTINGS.get('publisher_name'),
            "publisher_id": SETTINGS.get('publisher_id'),
            "publisher_name": SETTINGS.get('publisher_name'),
            "publisher_ipi_name": SETTINGS.get('publisher_ipi_name'),
            "publisher_ipi_base": SETTINGS.get('publisher_ipi_base', ''),
            "publisher_pr_society": SETTINGS.get(
                'publisher_pr_society'),
            "publisher_mr_society": SETTINGS.get(
                'publisher_mr_society'),
            "publisher_sr_society": SETTINGS.get(
                'publisher_sr_society'),
        })
        us = SETTINGS.get('us_publisher_override', {})
        if us:
            j["ascap_publisher"] = us.get('ASCAP')
            j["bmi_publisher"] = us.get('BMI')
            j["sesac_publisher"] = us.get('SESAC')
        works = OrderedDict()
        for work in qs:
            works.update(work.json)
        j.update({"works": works})
        return j

    @property
    def json(self):
        """Property that wraps :meth:`get_json`.

        Returns:
            dict: JSON-serializable data structure
        """
        return self.get_json()

    def get_cwr(self):
        """Get CWR from the external service, and save.

        Returns:
            None

        Raises:
            ValidationError: Validation errors are raised in case of
                network issues, bad requests, etc, so they can be
                handeled by admin.
        """
        if self.cwr:
            return
        url = SETTINGS.get('generator_url')
        json = self.json
        self._work_count = len(json.get('works'))
        if url is None:
            raise ValidationError('CWR generator URL not set.')
        try:
            response = requests.post(
                url,
                headers={'Authorization': 'Token {}'.format(
                    SETTINGS.get('token'))},
                json=json, timeout=30)
        except requests.exceptions.ConnectionError:
            raise ValidationError('Network error', code='service')
        if response.status_code == 400:
            raise ValidationError('Bad Request', code='service')
        elif response.status_code == 401:
            raise ValidationError('Unauthorized', code='service')
        elif response.status_code != 200:
            raise ValidationError('Unknown Error', code='service')

        if self.is_version_3:
            cwr = response.json()['cwr']
            self.cwr = cwr[0:157]
            self.year = self.cwr[56:58]
        else:
            self.cwr = response.json()['cwr']
            self.year = self.cwr[66:68]
        nr = CWRExport.objects.filter(year=self.year)
        nr = nr.order_by('-num_in_year').first()
        if nr:
            self.num_in_year = nr.num_in_year + 1
        else:
            self.num_in_year = 1
        if self.is_version_3:
            # self.filename depends on self.year !!!
            self.cwr += self.filename.ljust(27)
            self.cwr += cwr[184:]
        self.save()


class WorkAcknowledgement(models.Model):
    """Acknowledgement of work registration.

    Attributes:
        date (django.db.models.DateField): Acknowledgement date
        remote_work_id (django.db.models.CharField): Remote work ID
        society_code (django.db.models.CharField): 3-digit society code
        status (django.db.models.CharField): 2-letter status code
        TRANSACTION_STATUS_CHOICES (tuple): choices for status
        work (django.db.models.ForeignKey): FK to Work
    """

    class Meta:
        verbose_name = 'Registration Acknowledgement'

    TRANSACTION_STATUS_CHOICES = (
        ('CO', 'Conflict'),
        ('DU', 'Duplicate'),
        ('RA', 'Transaction Accepted'),
        ('AS', 'Registration Accepted'),
        ('AC', 'Registration Accepted with Changes'),
        ('RJ', 'Rejected'),
        ('NP', 'No Participation'),
        ('RC', 'Claim rejected'),
    )

    work = models.ForeignKey(Work, on_delete=models.PROTECT)
    society_code = models.CharField(
        'Society', max_length=3, choices=SOCIETIES)
    date = models.DateField()
    status = models.CharField(max_length=2, choices=TRANSACTION_STATUS_CHOICES)
    remote_work_id = models.CharField(max_length=20, blank=True)

    def __str__(self):
        return self.status


class ACKImport(models.Model):
    """CWR acknowledgement file import.

    Attributes:
        date (django.db.models.DateField): Acknowledgement date
        filename (django.db.models.CharField): Description
        report (django.db.models.CharField): Basically a log
        society_code (models.CharField): 3-digit society code,
            please note that ``choices`` is not set.
        society_name (models.CharField): Society name,
            used if society code is missing.
    """

    class Meta:
        verbose_name = 'CWR ACK Import'

    filename = models.CharField(max_length=60, editable=False)
    society_code = models.CharField(max_length=3, editable=False)
    society_name = models.CharField(max_length=45, editable=False)
    date = models.DateField(editable=False)
    report = models.TextField(editable=False)

    def __str__(self):
        return self.filename



# Generated by Django 2.1 on 2018-09-11 10:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0007_auto_20180911_1017'),
    ]

    operations = [
        migrations.AddField(
            model_name='writer',
            name='publisher_fee',
            field=models.DecimalField(blank=True, decimal_places=2, help_text='Percentage of royalties kept by the publisher', max_digits=5, null=True),
        ),
        migrations.AddField(
            model_name='writerinwork',
            name='publisher_fee',
            field=models.DecimalField(blank=True, decimal_places=2, help_text='Percentage of royalties kept by the publisher', max_digits=5, null=True),
        ),
        migrations.AlterField(
            model_name='writer',
            name='generally_controlled',
            field=models.BooleanField(default=False, verbose_name='General Agreement'),
        ),
    ]



# Generated by Django 2.1.3 on 2018-11-17 17:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0011_auto_20181114_1649'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='artistinwork',
            options={'verbose_name': 'Artist performing work', 'verbose_name_plural': 'Artists performing works'},
        ),
        migrations.AlterField(
            model_name='cwrexport',
            name='nwr_rev',
            field=models.CharField(choices=[('NWR', 'CWR 2.1: New work registration'), ('REV', 'CWR 2.1: Revision of registered work'), ('WRK', 'CWR 3.0: Work registration (alpha)')], db_index=True, default='NWR', max_length=3, verbose_name='NWR/REV'),
        ),
    ]



# Generated by Django 2.1.3 on 2018-11-14 16:49

from django.db import migrations, models
import django.db.models.deletion
import music_publisher.base


class Migration(migrations.Migration):

    dependencies = [
        ('music_publisher', '0010_auto_20181114_1548'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='firstrecording',
            name='catalog_number',
        ),
        migrations.AddField(
            model_name='firstrecording',
            name='artist',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='music_publisher.Artist', verbose_name='Recording Artist'),
        ),
        migrations.AddField(
            model_name='firstrecording',
            name='record_label',
            field=models.CharField(blank=True, default='FOO BAR MUSIC', max_length=60, validators=[music_publisher.base.CWRFieldValidator('first_album_label')]),
        ),
        migrations.AlterField(
            model_name='artist',
            name='isni',
            field=models.CharField(blank=True, max_length=16, null=True, unique=True, validators=[music_publisher.base.CWRFieldValidator('isni')], verbose_name='ISNI'),
        ),
    ]



"""Interface for :mod:`music_publisher` app.

Please note that this is the only interface.

Attributes:
    IS_POPUP_VAR (bool): :attr:`django.contrib.admin.options.IS_POPUP_VAR`
"""
from datetime import datetime
from decimal import Decimal
from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.db import models
from django.forms import ModelForm, FileField
from django.forms.models import BaseInlineFormSet
from django.http import HttpResponse, HttpResponseRedirect, JsonResponse
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
# from django.utils.decorators import method_decorator
from django.utils.html import mark_safe
from django.utils.timezone import now
# from django.views.decorators.csrf import csrf_protect
from .models import (
    AlbumCD, AlternateTitle, Artist, ArtistInWork, FirstRecording, Work,
    Writer, WriterInWork, VALIDATE, CWRExport, ACKImport, WorkAcknowledgement,
    WORK_ID_PREFIX)
import re
import requests


if hasattr(settings, 'MUSIC_PUBLISHER_SETTINGS'):
    SETTINGS = settings.MUSIC_PUBLISHER_SETTINGS
else:
    SETTINGS = {}

IS_POPUP_VAR = admin.options.IS_POPUP_VAR


class MusicPublisherAdmin(admin.ModelAdmin):
    """Custom Admin class, easily extendable.

    Should only be used for admin classes for top-level models."""

    save_as = True

    def save_model(self, request, obj, form, change):
        """Do super().save_model, msg if validation was not performed.
        """

        super().save_model(request, obj, form, change)
        if not VALIDATE:
            messages.add_message(
                request, messages.WARNING,
                mark_safe(
                    'Validation was not performed. Your data may be corrupt. '
                    'Please acquire a Validation and CWR generation licence '
                    'from <a href="https://matijakolaric.com/z_'
                    'contact/" target="_blank">matijakolaric.com</a>.'))
        elif not obj._cwr:
            messages.add_message(
                request, messages.ERROR,
                mark_safe(
                    'Validation failed, object saved, but marked as not'
                    'CWR-compliant. Please re-evaluate.'))


@admin.register(Writer)
class WriterAdmin(MusicPublisherAdmin):
    """Interface for :class:`.models.Writer`.
    """

    list_display = ('last_name', 'first_name', 'ipi_name', 'pr_society',
                    '_can_be_controlled', 'generally_controlled')

    def get_list_display(self, *args, **kwargs):
        """Return the list of fields based on settings.

        Original Publisher is important for the US."""

        lst = list(self.list_display)
        if SETTINGS.get('admin_show_publisher'):
            lst.append('original_publisher')
        lst.append('_cwr')
        return lst

    list_filter = ('_can_be_controlled', 'generally_controlled',
                   'pr_society', '_cwr')

    search_fields = ('last_name', 'ipi_name')
    readonly_fields = ('_can_be_controlled', 'original_publisher')
    fieldsets = (
        ('Name', {
            'fields': (
                ('first_name', 'last_name'),)}),
        ('IPI', {
            'fields': (
                ('ipi_name', 'ipi_base'),
                'pr_society'),
        }),
        ('General agreement', {
            'fields': (
                ('generally_controlled', ('saan', 'publisher_fee'))
                if (SETTINGS.get('admin_show_saan') or
                    SETTINGS.get('enforce_saan'))
                else ('generally_controlled', 'publisher_fee')),
        }),
    )
    actions = None

    def original_publisher(self, obj):
        """Return the original publisher.

        This makes sense only in the US context."""

        if obj.generally_controlled and obj.pr_society:
            return obj.get_publisher_dict().get('publisher_name')
        return ''

    def save_model(self, request, obj, form, *args, **kwargs):
        """Perform normal save_model, then update last_change of
        all connected works."""
        super().save_model(request, obj, form, *args, **kwargs)
        if form.changed_data:
            qs = Work.objects.filter(writerinwork__writer=obj)
            qs.update(last_change=now())


class AlternateTitleInline(admin.TabularInline):
    """Inline interface for :class:`.models.AlternateTitle`.
    """
    model = AlternateTitle
    extra = 0


class ArtistInWorkInline(admin.TabularInline):
    """Inline interface for :class:`.models.ArtistInWork`.
    """
    autocomplete_fields = ('artist', 'work')
    model = ArtistInWork
    extra = 0


class WriterInWorkFormSet(BaseInlineFormSet):
    """Formset for :class:`WriterInWorkInline`.
    """

    def clean(self):
        """Performs these checks:
            at least one writer must be controlled,
            sum of relative shares must be ~100%

        Returns:
            None

        Raises:
            ValidationError
        """
        super().clean()
        total = 0
        controlled = False
        for form in self.forms:
            if not form.is_valid():
                return
            if form.cleaned_data and not form.cleaned_data.get('DELETE'):
                total += form.cleaned_data['relative_share']
                if form.cleaned_data['controlled']:
                    controlled = True
        if not controlled:
            raise ValidationError('At least one writer must be controlled.')
        if not(Decimal(99.98) <= total <= Decimal(100.02)):
            raise ValidationError('Sum of relative shares must be 100%.')


class WriterInWorkInline(admin.TabularInline):
    """Inline interface for :class:`.models.WriterInWork`.
    """

    autocomplete_fields = ('writer', )
    readonly_fields = ('original_publisher',)
    model = WriterInWork
    formset = WriterInWorkFormSet
    extra = 0
    min_num = 1  # One writer is required
    fields = ('writer', 'capacity', 'relative_share', 'controlled')

    def get_fields(self, *args, **kwargs):
        """Return list of fields depending on settings.
        """
        lst = list(self.fields)
        if SETTINGS.get('admin_show_publisher'):
            lst.append('original_publisher')
        if SETTINGS.get('admin_show_saan') or SETTINGS.get('enforce_saan'):
            lst.append('saan')
        lst.append('publisher_fee')
        return lst

    def original_publisher(self, obj):
        """Return the original publisher.

        This makes sense only in the US context."""
        if obj.controlled and obj.writer:
            return obj.writer.get_publisher_dict().get('publisher_name')
        return ''


class RecordingInline(admin.StackedInline):
    """Inline interface for :class:`.models.FirstRecording`,
        used in :class:`WorkAdmin` and :class:`ArtistAdmin`.
    """
    autocomplete_fields = ('album_cd', 'artist', 'work')
    fieldsets = (
        (None, {
            'fields': (
                ('work', 'record_label', 'artist'),
                ('isrc', 'duration', 'release_date'),
                ('album_cd', )),
        }),
    )
    formfield_overrides = {
        models.TimeField: {'widget': forms.TimeInput},
    }
    verbose_name_plural = 'First Recordings'
    model = FirstRecording
    extra = 0


class TrackInline(admin.StackedInline):
    """Inline interface for :class:`.models.FirstRecording`,
        used in :class:`AlbumCDAdmin`.
    """

    autocomplete_fields = ('work', 'artist')
    fieldsets = (
        (None, {
            'fields': (
                ('work',),
                ('record_label', 'artist'),
                ('isrc', 'duration', 'release_date'),
            ),
        }),
    )
    model = FirstRecording
    formfield_overrides = {
        models.TimeField: {'widget': forms.TimeInput},
    }
    verbose_name_plural = 'Tracks'
    extra = 0


class WorkAcknowledgementInline(admin.TabularInline):
    """Inline interface for :class:`.models.WorkAcknowledgement`,
        used in :class:`WorkAdmin`.

        Please note that normal users should only have a 'view' permission.
    """

    model = WorkAcknowledgement
    extra = 0
    fields = ('date', 'society_code', 'remote_work_id', 'status')


@admin.register(Work)
class WorkAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.Work`.

    This is by far the most important part of the interface.

    Attributes:
        actions (tuple): batch actions used:
            :meth:`create_cwr`,
            :meth:`create_json`
        inlines (tuple): inlines used in change view:
            :class:`AlternateTitleInline`,
            :class:`WriterInWorkInline`,
            :class:`RecordingInline`,
            :class:`ArtistInWorkInline`,
            :class:`WorkAcknowledgementInline`,
    """

    inlines = (
        AlternateTitleInline, WriterInWorkInline,
        RecordingInline, ArtistInWorkInline,
        WorkAcknowledgementInline)

    def writer_last_names(self, obj):
        """This is a standard way how writers are shown in other apps."""

        return ' / '.join(
            writer.last_name.upper() for writer in obj.writers.all())
    writer_last_names.short_description = 'Writers\' last names'
    writer_last_names.admin_order_field = 'writers__last_name'

    def percentage_controlled(self, obj):
        """Controlled percentage
        (sum of relative shares for controlled writers)

        Please note that writers in work are already included in the queryset
        for other reasons, so no overhead except summing.
        """
        return sum(
            wiw.relative_share for wiw in obj.writerinwork_set.all()
            if wiw.controlled)
    percentage_controlled.short_description = '% controlled'

    def work_id(self, obj):
        return obj.work_id
    work_id.short_description = 'Work ID'
    work_id.admin_order_field = 'id'

    def album_cd(self, obj):
        if not (hasattr(obj, 'firstrecording') and obj.firstrecording):
            return None
        return obj.firstrecording.album_cd
    album_cd.short_description = 'Album / Library CD'
    album_cd.admin_order_field = 'firstrecording__album_cd'

    def isrc(self, obj):
        if not (hasattr(obj, 'firstrecording') and obj.firstrecording):
            return None
        return obj.firstrecording.isrc
    isrc.short_description = 'ISRC'
    isrc.admin_order_field = 'firstrecording__isrc'

    def cwr_export_count(self, obj):
        """Return the count of CWR exports with the link to the filtered
        changelist view for :class:`CWRExportAdmin`."""

        count = obj.cwr_exports__count
        url = reverse('admin:music_publisher_cwrexport_changelist')
        url += '?works__id__exact={}'.format(obj.id)
        return mark_safe('<a href="{}">{}</a>'.format(url, count))
    cwr_export_count.short_description = 'CWRs'
    cwr_export_count.admin_order_field = 'cwr_exports__count'

    def duration(self, obj):
        if not (hasattr(obj, 'firstrecording') and
                obj.firstrecording and obj.firstrecording.duration):
            return None
        return obj.firstrecording.duration.strftime('%H:%M:%S')
    duration.admin_order_field = 'firstrecording__duration'

    readonly_fields = (
        'writer_last_names', 'work_id', 'json', 'cwr_export_count')
    list_display = (
        'work_id', 'title', 'iswc', 'writer_last_names',
        'percentage_controlled', 'duration', 'isrc', 'album_cd',
        'cwr_export_count', '_cwr')

    def get_queryset(self, request):
        """Optimized queryset for changelist view.
        """
        qs = super().get_queryset(request)
        qs = qs.prefetch_related('writerinwork_set')
        qs = qs.prefetch_related('writers')
        qs = qs.prefetch_related('firstrecording__album_cd')
        qs = qs.annotate(models.Count('cwr_exports'))
        return qs

    class HasISWCListFilter(admin.SimpleListFilter):
        """Custom list filter on the presence of ISWC.
        """

        title = 'Has ISWC'
        parameter_name = 'has_iswc'

        def lookups(self, request, model_admin):
            """Simple Yes/No filter
            """
            return (
                ('Y', 'Yes'),
                ('N', 'No'),
            )

        def queryset(self, request, queryset):
            """Filter on presence of :attr:`.iswc`.
            """
            if self.value() == 'Y':
                return queryset.exclude(iswc__isnull=True)
            elif self.value() == 'N':
                return queryset.filter(iswc__isnull=True)

    class HasRecordingListFilter(admin.SimpleListFilter):
        """Custom list filter on the presence of first recording.
        """
        title = 'Has First Recording'
        parameter_name = 'has_rec'

        def lookups(self, request, model_admin):
            """Simple Yes/No filter
            """
            return (
                ('Y', 'Yes'),
                ('N', 'No'),
            )

        def queryset(self, request, queryset):
            """Filter on presence of :class:`.models.FirstRecording`.
            """
            if self.value() == 'Y':
                return queryset.exclude(firstrecording__isnull=True)
            elif self.value() == 'N':
                return queryset.filter(firstrecording__isnull=True)

    list_filter = (
        HasISWCListFilter,
        HasRecordingListFilter,
        ('firstrecording__album_cd', admin.RelatedOnlyFieldListFilter),
        '_cwr',
        'last_change',
    )

    search_fields = (
        'title', 'alternatetitle__title', 'writerinwork__writer__last_name',
        '^iswc', '^id')

    def get_search_results(self, request, queryset, search_term):
        """Deal with the situation term is work ID.
        """
        if search_term.isnumeric():
            search_term = search_term.lstrip('0')
        return super().get_search_results(request, queryset, search_term)

    fieldsets = (
        (None, {
            'fields': (
                'work_id',
                ('title', 'iswc'))}),)

    def save_model(self, request, obj, form, *args, **kwargs):
        if form.changed_data:
            obj.last_change = now()
        super().save_model(request, obj, form, *args, **kwargs)

    def create_cwr(self, request, qs):
        """Batch action that redirects to the add view for
        :class:`CWRExportAdmin` with selected works.
        """
        url = reverse('admin:music_publisher_cwrexport_add')
        ids = qs.values_list('id', flat=True)
        return HttpResponseRedirect(
            '{}?works={}'.format(url, ','.join(str(i) for i in ids)))
    create_cwr.short_description = 'Create CWR from selected works.'

    def create_json(self, request, qs):
        """Batch action that downloads a JSON file containing selected works.

        Returns:
            JsonResponse: JSON file with selected works
        """

        j = CWRExport().get_json(qs)
        response = JsonResponse(j, json_dumps_params={'indent': 4})
        name = '{}{}'.format(WORK_ID_PREFIX, datetime.now().toordinal())
        cd = 'attachment; filename="{}.json"'.format(name)
        response['Content-Disposition'] = cd
        return response
    create_json.short_description = 'Export selected works.'

    actions = (create_cwr, create_json)

    def get_actions(self, request):
        """Custom action disabling the default ``delete_selected``."""
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions

    def get_inline_instances(self, request, obj=None):
        """Limit inlines in popups."""
        instances = super().get_inline_instances(request)
        if IS_POPUP_VAR in request.GET or IS_POPUP_VAR in request.POST:
            return [i for i in instances if type(i) not in [
                ArtistInWorkInline, RecordingInline, WorkAcknowledgementInline
            ]]
        return instances


@admin.register(CWRExport)
class CWRExportAdmin(admin.ModelAdmin):
    """Admin interface for :class:`.models.CWRExport`.
    """

    actions = None

    def cwr_ready(self, obj):
        """Is CWR created?"""
        return obj.cwr != ''
    cwr_ready.boolean = True

    def work_count(self, obj):
        """Return the work count from the database field, or count them.
        (dealing with legacy)"""

        return obj._work_count or obj.works__count
    cwr_ready.boolean = True

    def get_highlighted_data(self, obj):
        """Get CWR highlighting from the external service.

        (available only for CWR 2.1)
        """
        if obj.is_version_3:
            return obj.cwr
        data = {'cwr': obj.cwr}
        try:
            response = requests.post(
                SETTINGS.get('highlighter_url'),
                headers={'Authorization': 'Token {}'.format(
                    SETTINGS.get('token'))},
                json=data, timeout=10)
            html = response.json()['html']
        except Exception:
            html = obj.cwr
        return html

    def view_link(self, obj):
        if obj.cwr:
            url = reverse(
                'admin:music_publisher_cwrexport_change', args=(obj.id,))
            url += '?preview=true'
            return mark_safe(
                '<a href="{}" target="_blank">View CWR</a>'.format(url))

    def download_link(self, obj):
        if obj.cwr:
            url = reverse(
                'admin:music_publisher_cwrexport_change', args=(obj.id,))
            url += '?download=true'
            return mark_safe('<a href="{}">Download</a>'.format(url))

    def get_queryset(self, request):
        """Optimized query with count of works in the export.
        """
        qs = super().get_queryset(request)
        qs = qs.annotate(models.Count('works'))
        return qs

    autocomplete_fields = ('works', )
    list_display = (
        'filename', 'nwr_rev', 'work_count', 'cwr_ready', 'view_link',
        'download_link')

    list_filter = ('nwr_rev', 'year')
    search_fields = ('works__title',)

    def get_readonly_fields(self, request, obj=None):
        """Read-only fields differ if CWR has been completed."""
        if obj and obj.cwr:
            return (
                'nwr_rev', 'works', 'filename', 'view_link', 'download_link')
        else:
            return ()

    def get_fields(self, request, obj=None):
        """Shown fields differ if CWR has been completed."""
        if obj and obj.cwr:
            return (
                'nwr_rev', 'works', 'filename', 'view_link', 'download_link')
        else:
            return ('nwr_rev', 'works')

    def save_model(self, request, obj, form, change):
        """Django splits the saving process int two parts, which does not
            work in this case, so this is simply passing the main object
            through to :meth:`save_related`.
        """
        if not (hasattr(self, 'obj') and self.obj.cwr):
            super().save_model(request, obj, form, change)
            self.obj = obj
        elif hasattr(self, 'obj'):
            del self.obj

    def save_related(self, request, form, formsets, change):
        """:meth:`save_model` passes the main object, which is needed to fetch
            CWR form the external service, but only after related objects are
            saved.
        """
        if hasattr(self, 'obj'):
            super().save_related(request, form, formsets, change)
            try:
                self.obj.get_cwr()
            except ValidationError as e:
                messages.add_message(
                    request, messages.ERROR,
                    mark_safe(
                        'CWR could not be fetched from external service. '
                        'The reason is "{}". Please try saving again later. '
                        'Currently saved as draft.'.format(str(e))))
            finally:
                del self.obj

    def has_delete_permission(self, request, obj=None, **kwargs):
        """If CWR has been created, it can no longer be deleted, as it may
        have been sent. This may change once the delivery is automated."""

        if obj and obj.cwr:
            return None
        return super().has_delete_permission(request, obj, **kwargs)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        """Normal change view with two sub-views defined by GET parameters:

        Parameters:
            preview: that returns the highlighted preview of CWR file,
            download: that downloads the CWR file."""
        obj = get_object_or_404(CWRExport, pk=object_id)
        if 'preview' in request.GET:
            html = self.get_highlighted_data(obj)
            return render(request, 'raw_cwr.html', {
                **self.admin_site.each_context(request),
                'lines': html,
                'title': obj.filename})
        elif 'download' in request.GET:
            response = HttpResponse(obj.cwr.encode().decode('latin1'))
            cd = 'attachment; filename="{}"'.format(obj.filename)
            response['Content-Disposition'] = cd
            return response
        extra_context = {
            'show_save': False,
        }
        if obj.cwr:
            extra_context.update({
                'save_as': False,
                'show_save_and_continue': False,
                'show_delete': False,
            })
        return super().change_view(
            request, object_id, form_url='', extra_context=extra_context)


class ACKImportForm(ModelForm):

    """Form used for CWR acknowledgement imports.

    Attributes:
        acknowledgement_file (FileField): Field for file upload
    """

    class Meta:
        model = ACKImport
        fields = ('acknowledgement_file',)

    acknowledgement_file = FileField()

    RE_HDR = re.compile(
        r'^HDR(?:SO|AA)[0 ]{6}([ \d][ \d]\d)(.{45})01\.10(\d{8})\d{6}(\d{8})')

    def clean(self):
        """Perform usual clean, then process the file, returning the content
            field as if it was the TextField."""
        super().clean()
        cd = self.cleaned_data
        ack = cd.get('acknowledgement_file')
        filename = ack.name
        if not (len(filename) in [18, 19] and filename[-5:].upper() != '.V21'):
            raise ValidationError('Wrong file name format.')
        self.cleaned_data['filename'] = filename
        content = ack.file.read().decode('latin1')
        match = re.match(self.RE_HDR, content)
        if not match:
            raise ValidationError('Incorrect CWR header')
        code, name, date1, date2 = match.groups()
        self.cleaned_data['society_code'] = code
        self.cleaned_data['society_name'] = name.strip()
        self.cleaned_data['date'] = datetime.strptime(
            max([date1, date2]), '%Y%m%d').date()
        self.cleaned_data['acknowledgement_file'] = content


@admin.register(ACKImport)
class ACKImportAdmin(admin.ModelAdmin):
    """Admin interface for :class:`.models.ACKImport`.
    """

    def get_form(self, request, obj=None, **kwargs):
        """Returns a custom form for new objects, default one for changes.
        """
        if obj is None:
            return ACKImportForm
        return super().get_form(request, obj, **kwargs)

    list_display = (
        'filename', 'society_code', 'society_name', 'date')
    list_filter = ('society_code', 'society_name')
    fields = readonly_fields = (
        'filename', 'society_code', 'society_name', 'date', 'report')

    add_fields = ('acknowledgement_file',)

    def get_fields(self, reqest, obj=None):
        """Return different fields for add vs change.
        """
        if obj:
            return self.fields
        return self.add_fields

    RE_ACK = re.compile(re.compile(
        r'(?<=\n)ACK.{106}(.{20})(.{20})(.{8})(.{2})', re.S))

    def process(self, request, society_code, file_content):
        """Create appropriate WorkAcknowledgement objects, without duplicates.

        Big part of this code should be moved to the model, left here because
        messaging is simpler.
        """

        unknown_work_ids = []
        existing_work_ids = []
        report = ''
        for x in re.findall(self.RE_ACK, file_content):
            work_id, remote_work_id, dat, status = x
            # work ID is numeric with an optional string
            work_id = work_id.strip()
            PREFIX = SETTINGS.get('work_id_prefix', '')
            if work_id.startswith(PREFIX):
                work_id = work_id[len(PREFIX):]
            if not work_id.isnumeric():
                unknown_work_ids.append(work_id)
                continue
            work_id = int(work_id)
            remote_work_id = remote_work_id.strip()
            dat = datetime.strptime(dat, '%Y%m%d').date()
            work = Work.objects.filter(id=work_id).first()
            if not work:
                messages.add_message(
                    request, messages.ERROR,
                    'Unknown work ID: {}'.format(work_id))
                continue
            wa, c = WorkAcknowledgement.objects.get_or_create(
                work_id=work_id,
                remote_work_id=remote_work_id,
                society_code=society_code,
                date=dat,
                status=status)
            if not c:
                existing_work_ids.append(str(work_id))
            report += '{} {} <{}>\n'.format(
                work.work_id, work.title, wa.get_status_display())
        if unknown_work_ids:
            messages.add_message(
                request, messages.ERROR,
                'Unknown work IDs: {}'.format(', '.join(unknown_work_ids)))
        if existing_work_ids:
            messages.add_message(
                request, messages.ERROR,
                'Data already exists for some or all works. '
                'Affected work IDs: {}'.format(', '.join(existing_work_ids)))
        return report

    def save_model(self, request, obj, form, change):
        """Custom save_model, it ignores changes, validates the for for new
            instances, if valid, it processes the file and, upon success,
            calls ``super().save_model``."""
        if change:
            return
        if form.is_valid():
            cd = form.cleaned_data
            obj.filename = cd['filename']
            obj.society_code = cd['society_code']
            obj.society_name = cd['society_name']
            obj.date = cd['date']
            # TODO move process() to model, and handle messages here
            obj.report = self.process(
                request, obj.society_code, cd['acknowledgement_file'])
            super().save_model(request, obj, form, change)

    def has_delete_permission(self, request, obj=None, *args, **kwargs):
        """Deleting ACK imports is a really bad idea.
        """
        return False


@admin.register(AlbumCD)
class AlbumCDAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.AlbumCD`.
    """

    readonly_fields = ('library',)
    inlines = [TrackInline]

    def get_fieldsets(self, request, obj=None):
        """Fields depend on settings.
        """
        fieldsets = (
            ('Library', {
                'fields': (
                    ('cd_identifier', 'library'),)
            }) if SETTINGS.get('library') else (
                'Library not set', {'fields': ()}),
            ('First Album', {
                'fields': (
                    ('album_title', 'album_label'),
                    ('ean', 'release_date'))
            }),
        )
        return fieldsets

    def label_not_set(self, obj=None):
        """Return the text if label is not set"""
        return 'NOT SET'
    label_not_set.short_description = 'Label'

    list_display = (
        '__str__',
    )

    def get_list_display(self, *args, **kwargs):
        """The list of fields depends on settings.
        """
        lst = list(self.list_display)
        if SETTINGS.get('label'):
            lst.append('album_label')
        lst.append('album_title')
        lst.append('release_date')
        lst.append('ean')
        if SETTINGS.get('library'):
            lst.append('library')
            lst.append('cd_identifier')
        lst.append('_cwr')
        return lst

    list_filter = ('_cwr',)

    search_fields = ('album_title', '^cd_identifier')
    actions = None

    def get_inline_instances(self, request, obj=None):
        """Limit inlines in popups."""
        if IS_POPUP_VAR in request.GET or IS_POPUP_VAR in request.POST:
            return []
        return super().get_inline_instances(request)

    def save_model(self, request, obj, form, *args, **kwargs):
        """Save, then update ``last_change`` of the corresponding works.
        """
        super().save_model(request, obj, form, *args, **kwargs)
        if form.changed_data:
            qs = Work.objects.filter(firstrecording__album_cd=obj)
            qs.update(last_change=now())


@admin.register(Artist)
class ArtistAdmin(MusicPublisherAdmin):
    """Admin interface for :class:`.models.Artist`.
    """

    list_display = ('last_or_band', 'first_name', 'isni', '_cwr')
    search_fields = ('last_name', 'isni',)
    list_filter = ('_cwr',)
    inlines = [RecordingInline, ArtistInWorkInline]

    fieldsets = (
        ('Name', {
            'fields': (
                ('first_name', 'last_name'),)}),
        ('ISNI', {
            'fields': (
                'isni',),
        }),
    )

    def get_inline_instances(self, request, obj=None):
        """Limit inlines in popups."""
        if IS_POPUP_VAR in request.GET or IS_POPUP_VAR in request.POST:
            return []
        return super().get_inline_instances(request)

    def last_or_band(self, obj):
        """Placeholder for :attr:`.models.Artist.last_name`."""
        return obj.last_name
    last_or_band.short_description = 'Last or band name'
    last_or_band.admin_order_field = 'last_name'

    actions = None

    def save_model(self, request, obj, form, *args, **kwargs):
        """Save, then update ``last_change`` of the corresponding works.
        """
        super().save_model(request, obj, form, *args, **kwargs)
        if form.changed_data:
            qs = Work.objects.filter(artistinwork__artist=obj)
            qs.update(last_change=now())
